// -*- c-basic-offset: 4; indent-tabs-mode: nil -*-

/* ***** BEGIN LICENSE BLOCK *****
Version: MPL 1.1/GPL 2.0/LGPL 2.1

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Standard-Sitemap Navigator code.

The Initial Developer of the Original Code is the
Standard-Sitemap Team.

Contributor(s): (see license.txt)

Alternatively, the contents of this file may be used under the terms of
either the GNU General Public License Version 2 or later (the "GPL"), or
the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
in which case the provisions of the GPL or the LGPL are applicable instead
of those above. If you wish to allow use of your version of this file only
under the terms of either the GPL or the LGPL, and not to allow others to
use your version of this file under the terms of the MPL, indicate your
decision by deleting the provisions above and replace them with the notice
and other provisions required by the GPL or the LGPL. If you do not delete
the provisions above, a recipient may use your version of this file under
the terms of any one of the MPL, the GPL or the LGPL.
***** END LICENSE BLOCK ***** */

Components.utils.import("resource://modules.standard-sitemap.org/sslib.jsm");


addEventListener
('load',
 function(event) {
     var obj = {
         lib : Application.storage.get("org.standard-sitemap.library", null),

         itemStorage : [],

         monitor : window["org.standard-sitemap.monitor"],

         siteUpdate : function(bReload, sReason) {
             var browser = gBrowser.selectedTab.linkedBrowser;
             var agent = this.monitor.getAgentForBrowser(browser);

             //this.lib.debug('Button updates for ' + sReason);
             var button = document.getElementById('sitemap-menu-button');
             var oSitemapURI = agent.getSitemapURI();
             if (oSitemapURI) {
                 //this.lib.debug('Awaiting content for ' + oSitemapURI.spec);
                 var self = this;
                 function cb(sm) {
                     if (self.updateItems(oSitemapURI, sm))
                         self.lib.xulEnable(button);
                 }
                 agent.getSitemap(cb);
             } else {
                 //this.lib.debug('Disabled button on ' + sReason);
                 this.lib.xulDisable(button);
             }
         },

         setDisplayLevel : function(val) {
             var browser = gBrowser.selectedTab.linkedBrowser;
             var agent = this.monitor.getAgentForBrowser(browser);
             agent.setDisplayLevel("menu", val);
         },

         popupShowing : function(event) {
             if (this.lib.checkLanguagePref("menu"))
                 this.siteUpdate(false, 'language-change');
             return true;
         },

         popupShown : function(event) {
             // Wierd.  We have to go through this business of
             // building the string s for the test to work reliably,
             // even though we don't ultimately print out s.
             var popup = document.getElementById('sitemap-popup');
             var s = "popupShown:";
             if (event.target == popup) s += " target";
             if (event.originalTarget == popup) s += " original";
             if (event.currentTarget == popup) s += " current";
             //this.lib.debug(s);
             if (event.originalTarget == popup) {
                 this.setDisplayLevel(45);
             }
         },

         popupHidden : function(event) {
             // Wierd.  We have to go through this business of
             // building the string s for the test to work reliably,
             // even though we don't ultimately print out s.
             var popup = document.getElementById('sitemap-popup');
             var s = "popupHidden:";
             if (event.target == popup) s += " target";
             if (event.originalTarget == popup) s += " original";
             if (event.currentTarget == popup) s += " current";
             //this.lib.debug(s);
             if (event.originalTarget == popup) {
                 this.setDisplayLevel(35);
             }
         },

         lastSitemap : null,
         lastLanguages : null,

         updateItems : function(oSitemapURI, oSitemap) {
             var lib = this.lib;
             //lib.debug('Item updates for ' + oSitemapURI.spec);

             var start = document.getElementById('sitemap-menu-sep');
             if (!start) {
                 this.setDisplayLevel(15);
                 return;
             }
             var rootmenu = start.parentNode;

             var newLanguages = lib.getLanguagePref(oSitemapURI);
             function sameLanguages(a, b) {
                 if (a == null)
                     return b == null;
                 if (b == null) return false;
                 if (b.length !=
                     a.length)
                     return false;
                 for (var i = 0; i < b.length; i++)
                     if (b[i] != a[i])
                         return false;
                 return true;
             }

             function makeFunc(spec, referMode, parentUiURI, uiURI) {
                 return function(event) {
                     event.stopPropagation();

                     // Determine the referrer URI.
                     var browser = gBrowser.selectedTab.linkedBrowser;
                     var currentURI = browser.currentURI;
                     var referrer =
                         lib.getReferrer(referMode,
                                         oSitemapURI, parentUiURI, uiURI,
                                         currentURI);
                     // lib.debug('\nReferrer: ' +
                     //           (referrer ? referrer.spec : null) +
                     //           '\nDestination: ' + spec);
                     window.top.openUILink(spec, event, false, false,
                                           false, null, referrer);
                     return false;
                 };
             }

             var bSameSitemap = oSitemap == this.lastSitemap;
             var bSameLanguages =
                 sameLanguages(newLanguages, this.lastLanguages);
             if (!bSameSitemap || !bSameLanguages) {
                 // If there's somewhere to store the previous items,
                 // do so.
                 if (this.lastSitemap) {
                     var storage = this.itemStorage[this.lastSitemap.id];
                     if (!storage)
                         this.itemStorage[this.lastSitemap.id] = storage = {};
                 } else {
                     // Create a dummy object, which we will discard.
                     var storage = {};

                     // TODO: Instead of a dummy object, just don't
                     // bother trying to store anything in it.
                 }

                 //lib.debug('Same sitemap: ' + bSameSitemap);
                 //lib.debug('Same languages: ' + bSameLanguages);

                 // Swap out the current menu items, storing them in
                 // the old sitemap object.
                 //lib.debug('Storing menu');
                 storage.oldMenuItems = [];
                 var nextElem = start.nextSibling;
                 for (var elem = nextElem; elem; elem = nextElem) {
                     nextElem = elem.nextSibling;
                     rootmenu.removeChild(elem);
                     storage.oldMenuItems.push(elem);
                 }

                 // Also store the language selection.
                 storage.oldMenuLanguages = this.lastLanguages;

                 // Get the storage for the new sitemap displayed in
                 // this window.
                 var storage = this.itemStorage[oSitemap.id];
                 if (!storage) {
                     //this.lib.debug('Nothing to recover');
                     storage = {};
                 }

                 if (storage.oldMenuItems &&
                     sameLanguages(storage.oldMenuLanguages, newLanguages)) {
                     // Try to recover menu items from the new sitemap.
                     //lib.debug('Restoring children');
                     for each (var elem in storage.oldMenuItems) {
                         rootmenu.appendChild(elem);
                     }
                 } else {
                     var key = "org.standard-sitemap.menu.node";
                     var doc = rootmenu.ownerDocument;
                     var visitText =
                         lib.TranslateString("sitemap.visit-here");
                     var bSepPending = false;

                     // Convert the sitemap into a menu structure.
                     function ensureSep(item) {
                         var lc = item.lastChild;
                         bSepPending = false;
                         if (!lc) return;
                         if (lc.localName == "menuseparator" &&
                             lc.namespaceURI == lib.xulNS)
                             return;
                         var ne = doc
                             .createElementNS(lib.xulNS, "menuseparator");
                         item.appendChild(ne);
                     }

                     function clearSep(item) {
                         if (!bSepPending) return;
                         bSepPending = false;
                         ensureSep(item);
                     }

                     function addHeading(sText, container) {
                         clearSep(container);
                         var item =
                             doc.createElementNS(lib.xulNS, "menuitem");
                         item.setAttribute('label', sText);
                         item.setAttribute('disabled', "true");
                         container.appendChild(item);
                     }

                     function addItem(node, sText, sDesc, uri, referMode,
                                      container, containerURI) {
                         // First determine whether this item has
                         // subitems, and we can display them without
                         // creating a loop.
                         var bHasChildren = node.aChildNodes &&
                             node.aChildNodes.length > 0;
                         for (var anc = container;
                              bHasChildren && anc && anc != rootmenu;
                              anc = anc.parentNode) {
                             if (anc.getUserData(key) == node)
                                 bHasChildren = false;
                         }

                         if (bHasChildren) {
                             // Create a <menu> with the textual details.
                             clearSep(container);
                             var item =
                                 doc.createElementNS(lib.xulNS, "menu");
                             item.setAttribute('label', sText);
                             item.setUserData(key, node, null);
                             container.appendChild(item);

                             // Create a <menupopup> to hold subitems.
                             var popup =
                                 doc.createElementNS(lib.xulNS, "menupopup");
                             item.appendChild(popup);
                             if (uri) {
                                 // If the parent item has a location,
                                 // make its first subitem be the
                                 // actual link.
                                 var me = doc
                                     .createElementNS(lib.xulNS, "menuitem");
                                 me.setAttribute("label", visitText);
                                 if (sDesc)
                                     me.setAttribute('description', sDesc);
                                 // TODO: Set oncommand instead?
                                 me.onclick = makeFunc(uri.spec, referMode,
                                                       containerURI,
                                                       node.getSitemapURI());
                                 me.setAttribute('statustext', uri.spec);
                                 popup.appendChild(me);
                                 ensureSep(popup);
                             }
                             processChildren(node.aChildNodes,
                                             node.autoGroupOrder, popup,
                                             uri ? uri : containerURI);
                             bSepPending = false;
                         } else {
                             // Create a single item.
                             clearSep(container);
                             var item =
                                 doc.createElementNS(lib.xulNS, "menuitem");
                             item.setAttribute('label', sText);
                             if (sDesc)
                                 item.setAttribute('description', sDesc);
                             if (uri) {
                                 // TODO: Set oncommand instead?
                                 item.onclick = makeFunc(uri.spec, referMode,
                                                         containerURI,
                                                         node.getSitemapURI());
                                 item.setAttribute('statustext', uri.spec);
                             }
                             container.appendChild(item);
                         }
                     }

                     var challenge = lib.getMenuLimitPref();

                     function getName(node) {
                         if (!node) return '';
                         var content = node.GetContent(newLanguages);
                         if (!content) {
                             content =
                                 node.GetContent(lib.GetAcceptLanguages());
                         }
                         if (!content) {
                             content = node.GetContent();
                         }
                         return content['name'];
                     }

                     function nameDiff(mode, n1, n2) {
                         var name1 = getName(n1);
                         var name2 = getName(n2);

                         switch (mode) {
                         case 'lexical':
                             for (var i = 0;
                                  i < name1.length && i < name2.length; i++) {
                                 if (name1[i].toUpperCase() !=
                                     name2[i].toUpperCase())
                                     return name1.substring(0, i + 1);
                             }
                             return name1.substring(0, i + 1);

                         default:
                             return name1;
                         }
                     }

                     function processChildrenSubset(nodes, mode, container,
                                                    containerURI,
                                                    lower, upper) {
                         var quant = upper - lower;

                         if (quant <= challenge) {
                             //lib.debug('quant ' + quant +
                             //        ' <= challenge ' + challenge);

                             // The menu is small enough.  Just add
                             // the nodes directly.
                             for (var i = lower; i < upper; i++)
                                 processNode(nodes[i], container,
                                             containerURI);
                             return;
                         }

                         //lib.debug('quant ' + quant +
                         //        ' > challenge ' + challenge);

                         // How big would groups have to be to prevent
                         // exceeding the challenge?  Keep on dividing
                         // until we get in range of the challenge.
                         // That will tell us the depth we'll have to
                         // go to.
                         var depth = 1;
                         var amount = quant;
                         while (amount > challenge) {
                             amount = Math.round(amount / challenge + 0.5);
                             depth++;
                             //lib.debug('amount ' + amount +
                             //'; depth ' + depth);
                         }

                         // Take the nth root of the total number of
                         // children, rounding up.
                         var subgroups = Math.pow(quant, 1 / depth);
                         subgroups = Math.round(subgroups + 0.5);

                         //lib.debug('subgroups ' + subgroups);

                         // How big should each group be?  There are
                         // two adjacent sizes.  lowSize would not be
                         // enough if the total is not a multiple of
                         // it.
                         var lowSize = quant / subgroups;
                         var highSize = Math.round(lowSize + 0.5);
                         lowSize = highSize - 1;

                         //lib.debug('each group ' + lowSize +
                         //        ' or ' + highSize);

                         // How many are we short if we just use
                         // lowSize?  And then where do we want to put
                         // the extra members?
                         var shortfall = quant - lowSize * subgroups;
                         var start =
                             Math.round((subgroups - shortfall) / 2) + 1;
                         var end = start + shortfall;

                         //lib.debug('start ' + start +
                         //        '; end ' + end);

                         // Compute the boundaries for each group.
                         var boundary = [ lower ];
                         for (var i = 1; i <= subgroups; i++) {
                             boundary[i] = boundary[i - 1] + lowSize;
                             if (i >= start && i < end)
                                 boundary[i]++;
                             //lib.debug('boundary[' + i +
                             //        '] = ' + boundary[i]);
                         }
                         if (boundary[subgroups] != upper)
                             lib.debug('boundary[subgroups] != upper: ' +
                                       upper);

                         // Create containers for each subgroup, and
                         // populate with a corresponding sublist of
                         // the nodes, recursively.
                         for (var i = 0; i < subgroups; i++) {
                             var start = boundary[i];
                             var end = boundary[i + 1];

                             // TODO: Work out name based on first and
                             // last items in the subgroup.
                             var outerStartNode =
                                 start > 0 ? nodes[start - 1] : null;
                             var outerEndNode =
                                 end < nodes.length ? nodes[end] : null;
                             var startName =
                                 nameDiff(mode, nodes[start], outerStartNode);
                             var endName =
                                 nameDiff(mode, nodes[end - 1], outerEndNode);
                             var name = startName + "\u2026" + endName;

                             // Create the container.
                             var group = doc.createElementNS(lib.xulNS, "menu");
                             group.setAttribute('label', name);
                             container.appendChild(group);

                             // We need a popup within the menu too.
                             var popup =
                                 doc.createElementNS(lib.xulNS, "menupopup");
                             group.appendChild(popup);

                             // Populate with child items.
                             processChildrenSubset(nodes, mode, popup,
                                                   containerURI,
                                                   start, end); 
                         }
                     }

                     function processChildren(nodes, order, item, itemURI) {
                         // Does the parent have @order=lexical?
                         if (order && order != 'none') {
                             //lib.debug('order = [' + order + ']');

                             // Do subgrouping.
                             processChildrenSubset(nodes, order, item, itemURI,
                                                   0, nodes.length);
                         } else {
                             //lib.debug('no order');

                             // Don't do any subgrouping.
                             for each (var childNode in nodes) {
                                 processNode(childNode, item, itemURI);
                             }
                         }
                     }

                     function processNode(node, parentItem, parentItemURI) {
                         var content = node.GetContent(newLanguages);
                         if (!content) {
                             content =
                                 node.GetContent(lib.GetAcceptLanguages());
                         }
                         if (!content) {
                             content = node.GetContent();
                         }
                         if (node.bIsGroup) {
                             ensureSep(parentItem);
                             var name = content['name'];
                             if (name && name != "")
                                 addHeading(name, parentItem);
                             processChildren(node.aChildNodes,
                                             node.autoGroupOrder,
                                             parentItem,
                                             parentItemURI);
                             bSepPending = true;
                         } else {
                             addItem(node, content['name'], content['desc'],
                                     content['addr'], content['refer'],
                                     parentItem, parentItemURI);
                         }
                     }

                     //lib.debug('Rebuilding menu');
                     processChildren(oSitemap.oSitemapNode.aChildNodes,
                                     null, rootmenu);
                 }

                 // Set the role items, or disable them.
                 var roles = {
                     'snhome-item' : oSitemap.oHome,
                     'contact-item' : oSitemap.oContact,
                     'contentinfo-item' : oSitemap.oContentInfo,
                     'search-item' : oSitemap.oSearchPage,
                 };
                 for (var itemId in roles) {
                     var item = document.getElementById(itemId);
                     lib.xulDisable(item);
                     var node = roles[itemId];
                     if (!node) continue;
                     var content = node.get(newLanguages);
                     if (!content)
                         content = node.get(lib.GetAcceptLanguages());
                     if (!content)
                         content = node.get();
                     if (!content['addr']) continue;
                     // TODO: Set oncommand instead?
                     item.onclick = makeFunc(content['addr'].spec,
                                             content['refer'],
                                             null, oSitemap.oSitemapURI);
                     item.setAttribute('statustext', content['addr'].spec);
                     lib.xulEnable(item);
                 }

                 // Record that we have the new sitemap, in the
                 // current language.
                 this.lastSitemap = oSitemap;
                 this.lastLanguages = newLanguages;
             }

             //lib.debug('Menu ready');

             //        TODO: Now do other stuff, like highlight the menu
             //        items for the current location.

             this.setDisplayLevel(35);
             return true;
         },

         tabOpen : function(tab) {
             var browser = getBrowser().getBrowserForTab(tab);
             if (!browser) {
                 this.lib.debug('No browser');
                 return;
             }
             var agent = this.monitor.getAgentForBrowser(browser);
             var self = this;
             var listener = function(r, s) {
                 self.siteUpdate(r, s);
             };
             browser.setUserData('org.standard-sitemap.menu.listener',
                                 listener, null);
             agent.addListener(listener);

             this.siteUpdate(false, 'tab-open');
         },

         tabClose : function(tab) {
             var browser = getBrowser().getBrowserForTab(tab);
             if (!browser) {
                 this.lib.debug('No browser');
                 return;
             }
             var agent = this.monitor.getAgentForBrowser(browser);
             var listener = browser
                 .getUserData('org.standard-sitemap.menu.listener');
             agent.removeListener(listener);

             this.siteUpdate(false, 'tab-close');
         },

         tabSelect : function(tab) {
             this.siteUpdate(false, 'tab-switch');
         },

         start : function() {
             var tabs = getBrowser().tabContainer;
             if (!tabs) return;

             // Make sure we register all existing tabs.
             for (var i = 0; i < tabs.itemCount; i++)
                 this.tabOpen(tabs.getItemAtIndex(i));

             var self = this;

             tabs.addEventListener('TabOpen',
                                   function(event) {
                                       self.tabOpen(event.target);
                                   }, false);
             tabs.addEventListener('TabClose',
                                   function(event) {
                                       self.tabClose(event.target);
                                   }, false);
             tabs.addEventListener('TabSelect',
                                   function(event) {
                                       self.tabSelect(event.target);
                                   }, false);

             this.siteUpdate(false, 'new-window');
         },

         end : function() {
         },
     };
     getBrowser().setUserData('org.standard-sitemap.menu-handler', obj, null);

     var pu = document.getElementById('sitemap-popup');
     pu.addEventListener('popupshowing',
                         function(event) {
                             obj.popupShowing(event);
                         }, false);
     pu.addEventListener('popupshown',
                         function(event) {
                             obj.popupShown(event);
                         }, false);
     pu.addEventListener('popuphidden',
                         function(event) {
                             obj.popupHidden(event);
                         }, false);

     obj.start();
 }, false);

addEventListener
('unload',
 function(event) {
     var obj = getBrowser().getUserData('org.standard-sitemap.menu-handler');
     obj.end();
 }, false);
