// -*- c-basic-offset: 4; indent-tabs-mode: nil -*-

/* ***** BEGIN LICENSE BLOCK *****
Version: MPL 1.1/GPL 2.0/LGPL 2.1

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Standard-Sitemap Navigator code.

The Initial Developer of the Original Code is the
Standard-Sitemap Team.

Contributor(s): (see license.txt)

Alternatively, the contents of this file may be used under the terms of
either the GNU General Public License Version 2 or later (the "GPL"), or
the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
in which case the provisions of the GPL or the LGPL are applicable instead
of those above. If you wish to allow use of your version of this file only
under the terms of either the GPL or the LGPL, and not to allow others to
use your version of this file under the terms of the MPL, indicate your
decision by deleting the provisions above and replace them with the notice
and other provisions required by the GPL or the LGPL. If you do not delete
the provisions above, a recipient may use your version of this file under
the terms of any one of the MPL, the GPL or the LGPL.
***** END LICENSE BLOCK ***** */

/* $Id: overlay.js 531 2011-10-02 08:34:12Z ssjools $
 *
 * Overlay code for the main browser window.
 */

// The plan for branch separate-monitor.

// The job of overlay.xul and overlay.js will be to inject the
// urlbar button/indicator, and to detect new tabs.  For each
// tab, it will obtain the sitemap agent with
// window["org.standard-sitemap.monitor"].getAgentForTab(tab), inject the
// sidebar into the tab, and make the sidebar get sitemap data from
// the agent and render it.  A listener also needs to be registered
// with the agent - whenever this is called, the sidebar must refetch
// data from the agent, and adjust its own state as necessary
// (e.g. highlight a different tree item, load a different sitemap, or
// auto-hide/reveal the sidebar).
//
// Information available from the agent includes:
//
// * The address of the current page in the tab.
//
// * The address of the sitemap specified by the current page or its
//   HTTP headers.
//
// * The sitemap data itself, loaded on demand.
//
// Notification will happen if the current page URI changes (even if
// just by #anchor reference), or if the sitemap URI changes.
//
// The sidebar will also tell the urlbar button whether it should
// display or not, and it will tell the agent whether it is displaying
// a sitemap or not.

Components.utils.import("resource://modules.standard-sitemap.org/sslib.jsm");

var standardSitemapSidebarOverlay = {
    lib : Application.storage.get("org.standard-sitemap.library", null),

    monitor : window["org.standard-sitemap.monitor"],

    registerTab : function(tab) {
        var oTabBrowser = getBrowser();
        var oBrowser = oTabBrowser.getBrowserForTab(tab);
        var oSelectedPanel = oBrowser.parentNode;
        // In FF3, we've found the xul:notification box.  In FF4, we
        // find a xul:stack, and have to go up one more element.  In
        // fact, we keep going up until we run out or find the ruddy
        // node!  TODO: Find a more robust way of doing this.
        while (oSelectedPanel != null &&
               (oSelectedPanel.localName != "notificationbox" ||
                oSelectedPanel.namespaceURI != this.lib.xulNS))
            oSelectedPanel = oSelectedPanel.parentNode;
        var sPanelID = oSelectedPanel.id;

        if (!oSelectedPanel.mStandardSitemapBox) {
            // TODO: Make this an error of some sort?
            this.lib.debug('Sitemap box not available in ' + sPanelID);
            return;
        }
        oSelectedPanel.mStandardSitemapBox.hidden = false;
        oSelectedPanel.mStandardSitemapSplitter.hidden = false;

        if (false)
            this.lib.debug('Sidebar overlay - New tab ' + sPanelID);

        // Allow the sidebar to specify whether it should be visible.
        var menuItem = document.getElementById("menuitem_viewSitemapSidebar");
        var fSetVisibility = function(bVisible) {
            if (tab == oTabBrowser.selectedTab) {
                // Set the menu-item tick correctly.
                menuItem.setAttribute("checked", bVisible);

            }
            oSelectedPanel.mStandardSitemapBox.hidden = !bVisible;
            oSelectedPanel.mStandardSitemapSplitter.hidden = !bVisible;
        };
        oBrowser.setUserData('org.standard-sitemap.sidebar.setVisibility',
                             fSetVisibility, null);


        // Keep the availability icons up-to-date.
        var self = this;
        var fSetAvailability = function(bAvailable) {
            if (tab == getBrowser().selectedTab)
                self.updateAvailabilityIcons(bAvailable);
        };
        oBrowser.setUserData('org.standard-sitemap.sidebar.setAvailability',
                             fSetAvailability, null);

        // Load the sidebar layout into the tab.
        if (true) {
            // TODO: Drop this section if the alternative is reliable.
            var source =
                "chrome://standard-sitemap.org/content/sidebar.xul#" + sPanelID;
        } else {
            // TODO: This section is not reliable for some versions
            // before FF3.5.3, e.g. 3.0.10.
            var setData = function() {
                var doc =
                    oSelectedPanel.mStandardSitemapBrowser.contentDocument;
                doc.setUserData('org.standard-sitemap.browser', oBrowser, null);
                doc.setUserData('org.standard-sitemap.id', sPanelID, null);
            };

            oSelectedPanel.mStandardSitemapBrowser
                .addEventListener('DOMContentLoaded', setData, false);
            var source ="chrome://standard-sitemap.org/content/sidebar.xul";
        }
        oSelectedPanel.mStandardSitemapBrowser.setAttribute("src", source);
    },

    fListener : null,

    deregisterTab : function(tab) {
        var agent = this.monitor.getAgentForTab(tab);
        // TODO: Erm, some cleaning up?
    },

    updateAvailabilityIconsForBrowser : function(oBrowser) {
        if (oBrowser != gBrowser.selectedBrowser)
            return;
        var agent = this.monitor.getAgentForBrowser(oBrowser);
        this.updateAvailabilityIconsForAgent(agent);
    },

    updateAvailabilityIconsForAgent : function(agent) {
        this.updateAvailabilityIcons(!!agent.getSitemapURI());
    },

    updateAvailabilityIcons : function(bAvailable) {
        // Display the urlbar icon if the tab's current page has a
        // sitemap, or the user expects to see it at all times.
        var bShowIcon = bAvailable || this.lib.getBoolPref('urlbaricon');
        var urlbarIcon = document.getElementById('sitemap-urlbar-button');
        urlbarIcon.setAttribute('hidden', !bShowIcon);
        urlbarIcon.setAttribute('tooltiptext',
                                this.lib.GetString(bShowIcon ?
                                                   "sitemap.hidesitemap" :
                                                   "sitemap.showsitemap",
                                                   document));

        // Grey out the customizable toolbar button if no sitemap is
        // available for the current tab.
        var button = document.getElementById('sitemap-button');
        if (button) button.setAttribute('sitemap', bAvailable);
    },

    /* Toggle the sidebar of the current tab. */
    toggleSidebar : function(bUserCommand, event) {
        if (event &&
            event.target.localName != "toolbarbutton" &&
            event.target.localName != "button" &&
            (event.target.localName != "menuitem" ||
             event.target.getAttribute("id") != "menuitem_viewSitemapSidebar"))
            return;
        var sidebar = getBrowser().selectedBrowser
            .getUserData('org.standard-sitemap.sidebar');
        sidebar.toggleVisibility(bUserCommand);
        return;
    },

    start : function() {
        var tabs = getBrowser().tabContainer;
        if (!tabs) return;

        // Make sure we register all existing tabs.
        for (var i = 0; i < tabs.itemCount; i++)
            this.registerTab(tabs.getItemAtIndex(i))
            ;

        var self = this;

        document.getElementById('sitemap-urlbar-button').
            addEventListener('onclick', function(event) {
                self.toggleSidebar(true, event);
            }, false);

        tabs.addEventListener('TabOpen', function(event) {
            self.registerTab(event.target);
        }, false);

        tabs.addEventListener('TabClose', function(event) {
            self.deregisterTab(event.target);
        }, false);

        tabs.addEventListener('TabSelect', function(event) {
            var browser = gBrowser.selectedTab.linkedBrowser;
            var agent = window["org.standard-sitemap.monitor"]
                .getAgentForBrowser(browser);
            self.updateAvailabilityIconsForAgent(agent);
        }, false);
    },

    end : function() {
    },
};


// Start things up when a window is created.
addEventListener('load',
                  function(event) {
                     standardSitemapSidebarOverlay.start();
                 }, false);
addEventListener('unload',
                  function(event) {
                     standardSitemapSidebarOverlay.end();
                 }, false);
